import numpy as np
import scipy.interpolate as interp
import scipy.integrate as integrate

import astropy.units as u

from . import detector
from . import binary
from .utils import make_quant


def Get_SNR_Matrix(source,instrument,var_x,sample_rate_x,var_y,sample_rate_y):
    """Calculates SNR Matrix

    Parameters
    ----------
    source : object
        Instance of a gravitational wave source class
    instrument : object
        Instance of a gravitational wave detector class
    var_x : str
        x-axis variable
    sample_rate_x : int
        Number of samples at which SNRMatrix is calculated corresponding to the x-axis variable
    var_y : str
        y-axis variable
    sample_rate_y : array
        samples at which SNRMatrix was calculated corresponding to the y-axis variable

    Returns
    -------
    sample_x : array
        samples at which SNRMatrix was calculated corresponding to the x-axis variable
    sample_y : array
        samples at which SNRMatrix was calculated corresponding to the y-axis variable
    SNRMatrix : array-like
        the sample_rate_y X sample_rate_x matrix at which the SNR was calculated corresponding to the particular x and y-axis variable choices

    Notes
    -----
    Uses the variable given and the data range to sample the space either logrithmically or linearly based on the
    selection of variables. Then it computes the SNR for each value.
    Returns the variable ranges used to calculate the SNR for each matrix, then returns the SNRs with size of the sample_yXsample_x

    """

    source.instrument = instrument
    #Get Samples for variables
    [sample_x,sample_y,recalculate_strain,recalculate_noise] = Get_Samples(source,instrument,var_x,sample_rate_x,var_y,sample_rate_y)

    sampleSize_x = len(sample_x)
    sampleSize_y = len(sample_y)
    SNRMatrix = np.zeros((sampleSize_x,sampleSize_y))

    for i in range(sampleSize_x):
        for j in range(sampleSize_y):

            if recalculate_noise == 'x':
                #Update Attribute (also updates dictionary)
                setattr(instrument,var_x,sample_x[i])
                setattr(source,var_y, sample_y[j])
            elif recalculate_noise == 'y':
                #Update Attribute (also updates dictionary)
                setattr(source,var_x,sample_x[i])
                setattr(instrument,var_y, sample_y[j])
            elif recalculate_noise == 'both':
                #Update Attribute (also updates dictionary)
                setattr(instrument,var_x,sample_x[i])
                setattr(instrument,var_y, sample_y[j])
            elif recalculate_noise == 'neither':
                #Update Attribute (also updates dictionary)
                setattr(source,var_x,sample_x[i])
                setattr(source,var_y, sample_y[j])

            if recalculate_noise != 'neither':
                #Recalculate noise curves if something is varied
                if hasattr(instrument,'fT'):
                    del instrument.fT
                if hasattr(instrument,'P_n_f'):
                    del instrument.P_n_f
                if isinstance(instrument,detector.PTA) and hasattr(instrument,'_sensitivitycurve'):
                    del instrument._sensitivitycurve

            source.Check_Freq_Evol()
            if source.ismono: #Monochromatic Source and not diff EOB SNR
                if hasattr(source,'h_gw'):
                    del source.h_gw
                SNRMatrix[j,i] = Calc_Mono_SNR(source,instrument)
            else: #Chirping Source
                if recalculate_strain == True: #If we need to calculate the waveform everytime
                    #Delete old PhenomD waveform
                    if hasattr(source,'_phenomD_f'):
                        del source._phenomD_f
                    if hasattr(source,'_phenomD_h'):
                        del source._phenomD_h
                if hasattr(source,'f'):
                    del source.f
                if hasattr(source,'h_f'):
                    del source.h_f
                SNRMatrix[j,i] = Calc_Chirp_SNR(source,instrument)

    return [sample_x,sample_y,SNRMatrix]

def Get_Samples(source,instrument,var_x,sample_rate_x,var_y,sample_rate_y):
    """Gets the x and y-axis samples

    Parameters
    ----------
    source : object
        Instance of a gravitational wave source class
    instrument : object
        Instance of a gravitational wave detector class
    var_x : str
        x-axis variable
    sample_rate_x : int
        Number of samples at which SNRMatrix is calculated corresponding to the x-axis variable
    var_y : str
        y-axis variable
    sample_rate_y : array
        samples at which SNRMatrix was calculated corresponding to the y-axis variable

    Returns
    -------
    sample_x : array
        samples at which SNRMatrix was calculated corresponding to the x-axis variable
    sample_y : array
        samples at which SNRMatrix was calculated corresponding to the y-axis variable

    Notes
    -----
        The function uses that to create a
        sample space for the variable either in linear space (for q,x1,or x2) or logspace
        for everything else.

    """
    sample_x = []
    sample_y = []
    recalculate_strain = False
    recalculate_noise = 'neither'

    if var_x in source.var_dict.keys():
        var_x_dict = source.var_dict[var_x]
    elif var_x in instrument.var_dict.keys():
        recalculate_noise = 'x'
        var_x_dict = instrument.var_dict[var_x]
    else:
        raise ValueError(var_x + ' is not a variable in the source nor the instrument.')

    if var_y in source.var_dict.keys():
        var_y_dict = source.var_dict[var_y]
    elif var_y in instrument.var_dict.keys():
        if recalculate_noise == 'x':
            recalculate_noise = 'both'
        else:
            recalculate_noise = 'y'
        var_y_dict = instrument.var_dict[var_y]
    else:
        raise ValueError(var_y + ' is not a variable in the source nor the instrument.')

    if var_x_dict['min'] != None and var_x_dict['max'] != None: #If the variable has non-None 'min',and 'max' dictionary attributes
        if var_x == 'q' or var_x == 'chi1' or var_x == 'chi2':
            #Sample in linear space for mass ratio and spins
            sample_x = np.linspace(var_x_dict['min'],var_x_dict['max'],sample_rate_x)
            recalculate_strain = True #Must recalculate the waveform at each point
        elif var_x == 'T_obs':
            #sample in linear space for instrument variables
            T_obs_min = make_quant(var_x_dict['min'],'s')
            T_obs_max = make_quant(var_x_dict['max'],'s')
            sample_x = np.linspace(T_obs_min.value,T_obs_max.value,sample_rate_x)
        else:
            #Sample in log space for any other variables
            #Need exception for astropy variables
            if isinstance(var_x_dict['min'],u.Quantity) and isinstance(var_x_dict['max'],u.Quantity):
                sample_x = np.logspace(np.log10(var_x_dict['min'].value),np.log10(var_x_dict['max'].value),sample_rate_x)
            else:
                sample_x = np.logspace(np.log10(var_x_dict['min']),np.log10(var_x_dict['max']),sample_rate_x)

    if var_y_dict['min'] != None and var_y_dict['max'] != None: #If the variable has non-None 'min',and 'max' dictionary attributes
        if var_y == 'q' or var_y == 'chi1' or var_y == 'chi2':
            #Sample in linear space for mass ratio and spins
            sample_y = np.linspace(var_y_dict['min'],var_y_dict['max'],sample_rate_y)
            recalculate_strain = True #Must recalculate the waveform at each point
        elif var_y == 'T_obs':
            #sample in linear space for instrument variables
            T_obs_min = make_quant(var_y_dict['min'],'s')
            T_obs_max = make_quant(var_y_dict['max'],'s')
            sample_y = np.linspace(T_obs_min.value,T_obs_max.value,sample_rate_y)
        else:
            #Sample in log space for any other variables
            #Need exception for astropy variables
            if isinstance(var_y_dict['min'],u.Quantity) and isinstance(var_y_dict['max'],u.Quantity):
                sample_y = np.logspace(np.log10(var_y_dict['min'].value),np.log10(var_y_dict['max'].value),sample_rate_y)
            else:
                sample_y = np.logspace(np.log10(var_y_dict['min']),np.log10(var_y_dict['max']),sample_rate_y)

    return sample_x,sample_y,recalculate_strain,recalculate_noise

def Calc_Mono_SNR(source,instrument):
    """Calculates the SNR for a monochromatic source

    Parameters
    ----------
    source : object
        Instance of a gravitational wave source class
    instrument : object
        Instance of a gravitational wave detector class

    """
    if isinstance(instrument,detector.PTA):
    	source.h_gw = binary.Get_Mono_Strain(source,instrument.f_opt,strain_const='Optimal')
    else:
    	source.h_gw = binary.Get_Mono_Strain(source,instrument.f_opt,strain_const='Averaged')
    indxfgw = np.abs(instrument.fT-instrument.f_opt).argmin()

    return source.h_gw*np.sqrt(instrument.T_obs.to('s')/instrument.S_n_f[indxfgw])

def Calc_Chirp_SNR(source,instrument):
    """Calculates the SNR for an evolving source

    Parameters
    ----------
    source : object
        Instance of a gravitational wave source class
    instrument : object
        Instance of a gravitational wave detector class

    Notes
    -----
    Uses an interpolated method to align waveform and instrument noise, then integrates
    over the overlapping region. See eqn 18 from Robson,Cornish,and Liu 2018 <https://arxiv.org/abs/1803.01944>
    Values outside of the sensitivity curve are arbitrarily set to 1e30 so the SNR is effectively 0

    """

    #Use to integrate from initial observed frequency f(t_init) to f(t_init-T_obs)
    #Does not work unless t_init is randomly sampled, which we don't do
    #indxfgw_start = np.abs(source.f-source.f_init).argmin()
    #indxfgw_end = np.abs(source.f-source.f_T_obs).argmin()

    #Only want to integrate from observed frequency (f(T_obs_before_merger)) till merger
    indxfgw_start = np.abs(source.f-source.f_T_obs).argmin()
    indxfgw_end = len(source.f)
    if indxfgw_start >= len(source.f)-1:
        #If the SMBH has already merged set the SNR to ~0
        return 1e-30
    else:
        f_cut = source.f[indxfgw_start:indxfgw_end]
        h_cut = source.h_f[indxfgw_start:indxfgw_end]

    #################################
    #Interpolate the Strain Noise Spectral Density to only the frequencies the
    #strain runs over
    #Set Noise to 1e30 outside of signal frequencies
    S_n_f_interp_old = interp.interp1d(np.log10(instrument.fT.value),np.log10(instrument.S_n_f.value),kind='cubic',fill_value=30.0, bounds_error=False)
    S_n_f_interp_new = S_n_f_interp_old(np.log10(f_cut.value))
    S_n_f_interp = 10**S_n_f_interp_new

    if isinstance(instrument,detector.PTA):
        #Rescaled by 1.5 to make SNR plots match...
        integral_consts = 4.*1.5

    elif isinstance(instrument,detector.SpaceBased) or isinstance(instrument,detector.GroundBased):
        integral_consts = 16./5.


    #CALCULATE SNR FOR BOTH NOISE CURVES
    denom = S_n_f_interp #Sky Averaged Noise Spectral Density
    numer = h_cut**2
    integrand = numer/denom
    if isinstance(integrand,u.Quantity) and isinstance(f_cut,u.Quantity):
        SNRsqrd = integral_consts*np.trapz(integrand.value,f_cut.value,axis=0) #SNR**2
    else:
        SNRsqrd = integral_consts*np.trapz(integrand,f_cut,axis=0) #SNR**2

    return np.sqrt(SNRsqrd)


def saveSNR(sample_x,sample_y,SNRMatrix,save_location,SNR_filename,sample_filename):
    """Saves SNR Matrix

    Parameters
    ----------
    sample_x : array
        samples at which SNRMatrix was calculated corresponding to the x-axis variable
    sample_y : array
        samples at which SNRMatrix was calculated corresponding to the y-axis variable
    SNRMatrix : array-like
        the matrix at which the SNR was calculated corresponding to the particular x and y-axis variable choices
    save_location : str
        the directory to which the Samples and SNR are saved
    SNR_filename : str
        the name of the SNR file
    sample_filename : str
        the name of the sample file

    """
    np.savetxt(save_location+SNR_filename,SNRMatrix)
    np.savetxt(save_location+sample_filename,[sample_x,sample_y])



import numpy as np
import os
import astropy.constants as const
import astropy.units as u
import scipy.interpolate as interp
from astropy.cosmology import z_at_value
from astropy.cosmology import WMAP9 as cosmo

from . import utils

import hasasia.sensitivity as hassens
import hasasia.sim as hassim

current_path = os.getcwd()
splt_path = current_path.split("/")
top_path_idx = splt_path.index('gwent')
top_directory = "/".join(splt_path[0:top_path_idx+1])

class PTA:
    """
    Class to make a PTA instrument using the methods of Hazboun, Romano, Smith (2019) <https://arxiv.org/abs/1907.04341>

    Parameters
    ----------

    name : string
        name of the instrument

    T_obs : float
        the observation time of the PTA in [years]
    N_p : int
        the number of pulsars in the PTA
    sigma : float
        the rms error on the pulsar TOAs in [sec]
    cadence : float
        How often the pulsars are observed in [num/year]

    load_location : string, optional
        If you want to load a PTA curve from a file, it's the file path
    A_GWB : float, optional
        Amplitude of the gravitational wave background added as red noise
    alpha_GWB : float, optional
        the GWB power law, if empty and A_GWB is set, it is assumed to be -2/3
    A_rn : float, optional
        Individual pulsar red noise amplitude, is a list of [min,max] values from which to uniformly sample
    alpha_rn : float, optional
        Individual pulsar red noise alpha (power law), is a list of [min,max] values from which to uniformly sample
    f_low : float, optional
        Assigned lowest frequency of PTA (default assigns 1/(5*T_obs))
    f_high : float, optional
        Assigned highest frequency of PTA (default is Nyquist freq cadence/2)
    nfreqs : int, optional
        Number of frequencies in logspace the sensitivity is calculated

    """
    def __init__(self,name,*args,**kwargs):
        self.name = name
        for keys,value in kwargs.items():
            if keys == 'load_location':
                self.Load_Data(value)
            elif keys == 'A_GWB':
                self.A_GWB = value
            elif keys == 'alpha_GWB':
                self.alpha_GWB = value
            elif keys == 'A_rn':
                self.A_rn_min = value[0]
                self.A_rn_max = value[1]
            elif keys == 'alpha_rn':
                self.alpha_rn_min = value[0]
                self.alpha_rn_max = value[1]
            elif keys == 'f_low':
                self.f_low = utils.make_quant(value,'Hz')
            elif keys == 'f_high':
                self.f_high = utils.make_quant(value,'Hz')
            elif keys == 'nfreqs':
                self.nfreqs = value

        if not hasattr(self,'nfreqs'):
            self.nfreqs = int(1e3)
        if hasattr(self,'f_low') and hasattr(self,'f_high'):
            self.fT = np.logspace(np.log10(self.f_low.value),np.log10(self.f_high.value),self.nfreqs)

        if len(args) != 0:
            [T_obs,N_p,sigma,cadence] = args
            self.T_obs = utils.make_quant(T_obs,'yr')
            self.N_p = N_p
            self.sigma = utils.make_quant(sigma,'s')
            self.cadence = utils.make_quant(cadence,'1/yr')

    @property
    def T_obs(self):
        return self._T_obs
    @T_obs.setter
    def T_obs(self,value):
        self.var_dict = ['T_obs',value]
        self._T_obs = self._return_value

    @property
    def N_p(self):
        return self._N_p
    @N_p.setter
    def N_p(self,value):
        self.var_dict = ['N_p',value]
        self._N_p = self._return_value

    @property
    def cadence(self):
        return self._cadence
    @cadence.setter
    def cadence(self,value):
        self.var_dict = ['cadence',value]
        self._cadence = self._return_value

    @property
    def sigma(self):
        self._sigma = utils.make_quant(self._sigma,'s')
        return self._sigma
    @sigma.setter
    def sigma(self,value):
        self.var_dict = ['sigma',value]
        self._sigma = self._return_value

    @property
    def var_dict(self):
        return self._var_dict
    @var_dict.setter
    def var_dict(self,value):
        utils.Get_Var_Dict(self,value)

    @property
    def fT(self):
        if not hasattr(self,'_fT'):
            #frequency sampled from 1/observation time to nyquist frequency (c/2)
            #5 is the default value for now (from Hazboun et al. 2019)
            T_obs_sec = self.T_obs.to('s').value
            cadence_sec = self.cadence.to('1/s').value
            self._fT = np.logspace(np.log10(1/(5*T_obs_sec)),np.log10(cadence_sec/2),self.nfreqs)*u.Hz
        return self._fT
    @fT.setter
    def fT(self,value):
        self._fT = value
    @fT.deleter
    def fT(self):
        del self._fT

    @property
    def h_n_f(self):
        """Effective Strain Noise Amplitude"""
        if not hasattr(self,'_h_n_f'):
            if not hasattr(self,'_sensitivitycurve'):
                self.Init_PTA()
            self._h_n_f = self._sensitivitycurve.h_c
        return self._h_n_f
    @h_n_f.setter
    def h_n_f(self,value):
        self._h_n_f = value
    @h_n_f.deleter
    def h_n_f(self):
        del self._h_n_f

    @property
    def S_n_f(self):
        #Effective noise power amplitude
        if not hasattr(self,'_S_n_f'):
            if not hasattr(self,'_sensitivitycurve'):
                self.Init_PTA()
            self._S_n_f = self._sensitivitycurve.S_eff
            self._S_n_f = utils.make_quant(self._S_n_f,'1/Hz')
        return self._S_n_f
    @S_n_f.setter
    def S_n_f(self,value):
        self._S_n_f = value
    @S_n_f.deleter
    def S_n_f(self):
        del self._S_n_f

    @property
    def f_opt(self):
        #The optimal frequency of the instrument ie. the frequecy at the lowest strain
        if not hasattr(self,'_f_opt'):
            self._f_opt = self.fT[np.argmin(self.h_n_f)]
        return self._f_opt

    def Load_Data(self,load_location):
        self._I_data = np.loadtxt(load_location)
        self.fT = self._I_data[:,0]
        self.h_n_f = self._I_data[:,1]

    def Init_PTA(self):
        """Initializes a PTA in hasasia

        Notes
        -----
        See Hazboun, Romano, Smith (2019) <https://arxiv.org/abs/1907.04341> for details

        """

        #Random Sky Locations of Pulsars
        phi = np.random.uniform(0, 2*np.pi,size=self.N_p)
        cos_theta = np.random.uniform(-1,1,size=self.N_p)
        theta = np.arccos(cos_theta)

        if hasattr(self,'A_GWB'):
            if not hasattr(self,'alpha_GWB'):
                self.alpha_GWB = -2/3.
            #Make a set of psrs with the same parameters with a GWB as red noise
            psrs = hassim.sim_pta(timespan=self.T_obs.value,cad=self.cadence.value,sigma=self.sigma.value,\
                phi=phi, theta=theta, Npsrs=self.N_p,A_rn=self.A_GWB,alpha=self.alpha_GWB,freqs=self.fT.value)
        elif hasattr(self,'A_rn_min') or hasattr(self,'alpha_rn_min'):
            if not hasattr(self,'A_rn_min'):
                A_rn = np.random.uniform(1e-16,1e-12,size=self.N_p)
            else:
                A_rn = np.random.uniform(self.A_rn_min,self.A_rn_max,size=self.N_p)
            if not hasattr(self,'alpha_rn_min'):
                alphas = np.random.uniform(-3/4,1,size=self.N_p)
            else:
                alphas = np.random.uniform(self.alpha_rn_min,self.alpha_rn_max,size=self.N_p)
            #Make a set of psrs with uniformly sampled red noise
            psrs = hassim.sim_pta(timespan=self.T_obs.value,cad=self.cadence.value,sigma=self.sigma.value,\
                phi=phi, theta=theta, Npsrs=self.N_p,A_rn=A_rn,alpha=alphas,freqs=self.fT.value)
        else:
            #Make a set of psrs with the same parameters
            psrs = hassim.sim_pta(timespan=self.T_obs.value,cad=self.cadence.value,sigma=self.sigma.value,\
                phi=phi, theta=theta, Npsrs=self.N_p,freqs=self.fT.value)
        #Get Spectra of pulsars
        spectra= []
        for p in psrs:
             sp = hassens.Spectrum(p,freqs=self.fT.value)
             spectra.append(sp)

        self._sensitivitycurve = hassens.DeterSensitivityCurve(spectra)



class Interferometer:
    """
    Class to make an interferometer

    Parameters
    ----------

    name : string
        name of the instrument

    T_obs : float
        the observation time of the PTA in [years]

    load_location : string, optional
        If you want to load an instrument curve from a file, it's the file path
    I_type : string, optional
        Type of input data; can be the effective strain spectral density $S_{n}(f)$ ('ENSD'),
        the amplitude spectral density, $\sqrt{S_{n}(f)}$ ('ASD'), or the characteristic strain $h_{n}(f)$ ('h')

    """
    def __init__(self,name,T_obs,**kwargs):
        self.name = name
        self.T_obs = T_obs

        for keys,value in kwargs.items():
            if keys == 'load_location':
                self.load_location = value
            elif keys == 'I_type':
                self.I_type = value

        if hasattr(self,'load_location'):
            self.Load_Data()

    @property
    def T_obs(self):
        self._T_obs = utils.make_quant(self._T_obs,'yr')
        return self._T_obs
    @T_obs.setter
    def T_obs(self,value):
        self.var_dict = ['T_obs',value]
        self._T_obs = self._return_value

    @property
    def var_dict(self):
        return self._var_dict
    @var_dict.setter
    def var_dict(self,value):
        utils.Get_Var_Dict(self,value)

    @property
    def fT(self):
        if not hasattr(self,'_fT'):
            raise NotImplementedError('Interferometer frequency must be defined inside SpaceBased or GroundBased classes.')
        return self._fT
    @fT.setter
    def fT(self,value):
        self._fT = value

    @property
    def f_opt(self):
        """The optimal frequency of the instrument ie. the frequecy at the lowest strain"""
        self._f_opt = self.fT[np.argmin(self.h_n_f)]
        return self._f_opt

    @property
    def P_n_f(self):
        """Strain power sensitivity. """
        raise NotImplementedError('Power Spectral Density method must be defined inside SpaceBased or GroundBased classes.')

    @property
    def S_n_f(self):
        """Effective Noise Power Specral Density"""
        if not hasattr(self,'_S_n_f'):
            if hasattr(self,'_I_data'):
                if self._I_Type == 'ASD':
                    S_n_f_sqrt = self._I_data[:,1]
                    self._S_n_f = S_n_f_sqrt**2/u.Hz
                elif self._I_Type == 'ENSD':
                    self._S_n_f = self._I_data[:,1]/u.Hz
                elif self._I_Type == 'h':
                    self._S_n_f = self.h_n_f**2/self.fT
            else:
                raise NotImplementedError('Effective Noise Power Spectral Density method must be defined inside SpaceBased or GroundBased classes.')
        return self._S_n_f

    @property
    def h_n_f(self):
        """Characteristic Strain/effective strain noise amplitude"""
        if not hasattr(self,'_h_n_f'):
            if hasattr(self,'_I_data') and self._I_Type == 'h':
                self._h_n_f = self._I_data[:,1]
            else:
                self._h_n_f = np.sqrt(self.fT*self.S_n_f)
        return self._h_n_f

    def Load_Data(self):
        if not hasattr(self,'I_type'):
            print('Is the data:')
            print(' *Effective Noise Spectral Density - "E"')
            print(' *Amplitude Spectral Density- "A"')
            print(' *Effective Strain - "h"')
            self.I_type = input('Please enter one of the answers in quotations: ')
            self.Load_Data()

        if self.I_type == 'E' or self.I_type == 'e':
            self._I_Type = 'ENSD'
        elif self.I_type == 'A' or self.I_type == 'a':
            self._I_Type = 'ASD'
        elif self.I_type == 'h' or self.I_type == 'H':
            self._I_Type = 'h'
        else:
            print('Is the data:')
            print(' *Effective Noise Spectral Density - "E"')
            print(' *Amplitude Spectral Density- "A"')
            print(' *Effective Strain - "h"')
            self.I_type = input('Please enter one of the answers in quotations: ')
            self.Load_Data()

        self._I_data = np.loadtxt(self.load_location)
        self.fT = self._I_data[:,0]
        self.fT = utils.make_quant(self.fT,'Hz')


class GroundBased(Interferometer):
    """
    Class to make a Ground Based Instrument using the Interferometer base class
    """
    def __init__(self,name,T_obs,**kwargs):
        super().__init__(name,T_obs,**kwargs)
        """
        Currently doesn't do anything differently that Instrument object, can be updated if we ever construct at Ground Based PSD...
        """

    @property
    def P_n_f(self):
        """Power Spectral Density. """
        raise NotImplementedError('Power Spectral Density method must be defined inside SpaceBased or GroundBased classes.')
    @P_n_f.deleter
    def P_n_f(self):
        del self._P_n_f



class SpaceBased(Interferometer):
    """
    Class to make a Space Based interferometer

    Parameters
    ----------
    L : float
        the armlength the of detector in [meters]
    A_acc : float
        the Amplitude of the Acceleration Noise in [meters/second^2]
    f_acc_break_low : float
        the lower break frequency of the acceleration noise in [Hz]
    f_acc_break_high : float
        the higher break frequency of the acceleration noise in [Hz]
    A_IFO : float
        the amplitude of the interferometer

    T_Function_Type : string, optional
        Picks the transfer function generation method
        To use the numerically approximated method in Robson, Cornish, and Liu, 2019, input "N"
        To use the analytic fit in Larson, Hiscock, and Hellings, 2000, input "A"
    Background : Boolean
        Add in a Galactic Binary Confusion Noise
    f_low : float
        Assigned lowest frequency of instrument (default assigns 10^-5Hz)
    f_high : float
        Assigned highest frequency of instrument (default is 1Hz)
    nfreqs : int
        Number of frequencies in logspace the sensitivity is calculated (default is 1e3)

    """
    def __init__(self,name,T_obs,*args,**kwargs):
        super().__init__(name,T_obs,**kwargs)
        self.name = name
        for keys,value in kwargs.items():
            if keys == 'T_Function_Type':
                self.Set_T_Function_Type(value)
            elif keys == 'Background':
                self.Background = value
            elif keys == 'f_low':
                self.f_low = value
            elif keys == 'f_high':
                self.f_high = value
            elif keys == 'nfreqs':
                self.nfreqs = value

        if not hasattr(self,'nfreqs'):
            self.nfreqs = int(1e3)
        if not hasattr(self,'f_low'):
            self.f_low = 1e-5*u.Hz
        if not hasattr(self,'f_high'):
            self.f_high = 1.0*u.Hz
        if not hasattr(self,'Background'):
            self.Background = False

        if len(args) != 0:
            [L,A_acc,f_acc_break_low,f_acc_break_high,A_IFO,f_IMS_break] = args
            self.L = utils.make_quant(L,'m')
            self.A_acc = utils.make_quant(A_acc,'m/(s*s)')
            self.f_acc_break_low = utils.make_quant(f_acc_break_low,'Hz')
            self.f_acc_break_high = utils.make_quant(f_acc_break_high,'Hz')
            self.A_IFO = utils.make_quant(A_IFO,'m')
            self.f_IMS_break = utils.make_quant(f_IMS_break,'Hz')

        if not hasattr(self,'_T_Function_Type') and not hasattr(self,'load_location'):
            self.Set_T_Function_Type('N')

    @property
    def L(self):
        return self._L
    @L.setter
    def L(self,value):
        self.var_dict = ['L',value]
        self._L = self._return_value

    @property
    def A_acc(self):
        return self._A_acc
    @A_acc.setter
    def A_acc(self,value):
        self.var_dict = ['A_acc',value]
        self._A_acc = self._return_value

    @property
    def f_acc_break_low(self):
        return self._f_acc_break_low
    @f_acc_break_low.setter
    def f_acc_break_low(self,value):
        self.var_dict = ['f_acc_break_low',value]
        self._f_acc_break_low = self._return_value

    @property
    def f_acc_break_high(self):
        return self._f_acc_break_high
    @f_acc_break_high.setter
    def f_acc_break_high(self,value):
        self.var_dict = ['f_acc_break_high',value]
        self._f_acc_break_high = self._return_value

    @property
    def A_IFO(self):
        return self._A_IFO
    @A_IFO.setter
    def A_IFO(self,value):
        self.var_dict = ['A_IFO',value]
        self._A_IFO = self._return_value

    @property
    def f_IMS_break(self):
        return self._f_IMS_break
    @f_IMS_break.setter
    def f_IMS_break(self,value):
        self.var_dict = ['f_IMS_break',value]
        self._f_IMS_break = self._return_value

    @property
    def P_n_f(self):
        """Power Spectral Density"""
        if not hasattr(self,'_P_n_f'):
            if not hasattr(self,'_T_Function_Type'):
                self.Set_T_Function_Type()

            P_acc = self.A_acc**2*(1+(self.f_acc_break_low/self.fT)**2)*(1+(self.fT/(self.f_acc_break_high))**4)/(2*np.pi*self.fT)**4 #Acceleration Noise
            P_IMS = self.A_IFO**2*(1+(self.f_IMS_break/self.fT)**4) #Displacement noise of the interferometric TM--to-TM

            f_trans = const.c/2/np.pi/self.L #Transfer frequency
            self._P_n_f = (P_IMS + 2*(1+np.cos(self.fT.value/f_trans.value)**2)*P_acc)/self.L**2/u.Hz
        return self._P_n_f
    @P_n_f.deleter
    def P_n_f(self):
        del self._P_n_f

    @property
    def S_n_f(self):
        """Effective Noise Power Specral Density"""
        if not hasattr(self,'_S_n_f'):
            if hasattr(self,'_I_data'):
                if self._I_Type == 'ASD':
                    S_n_f_sqrt = self._I_data[:,1]
                    self._S_n_f = S_n_f_sqrt**2/u.Hz
                elif self._I_Type == 'ENSD':
                    self._S_n_f = self._I_data[:,1]/u.Hz
                elif self._I_Type == 'h':
                    self._S_n_f = self.h_n_f**2/self.fT
            else:
                S_n_f = self.P_n_f/self.transferfunction**2
                if self.Background:
                    self._S_n_f= S_n_f+self.Add_Background()
                else:
                    self._S_n_f = S_n_f
        return self._S_n_f
    @S_n_f.deleter
    def S_n_f(self):
        del self._S_n_f

    def Load_Transfer_Function(self):
        #Numerical transfer function
        Numerical_Transfer_Function_filedirectory = top_directory + '/docs/LoadFiles/NumericalTransferFunction/transfer.dat'
        Numerical_Transfer_Function_data = np.loadtxt(Numerical_Transfer_Function_filedirectory)
        self._transferfunctiondata = Numerical_Transfer_Function_data

    def Get_Numeric_Transfer_Function(self):
        if not hasattr(self,'_transferfunctiondata'):
            self.Load_Transfer_Function()

        fc = const.c/(2*self.L)  #light round trip freq
        LISA_Transfer_Function_f = fc*self._transferfunctiondata[:,0]

        idx_f_5 = np.abs(LISA_Transfer_Function_f-self.f_low).argmin()
        idx_f_1 = np.abs(LISA_Transfer_Function_f-self.f_high).argmin()

        #3/10 is normalization 2/5sin(openingangle)
        #Some papers use 3/20, not summing over 2 independent low-freq data channels
        self.transferfunction = np.sqrt(3/10)*self._transferfunctiondata[idx_f_5:idx_f_1,1]
        self.fT = LISA_Transfer_Function_f[idx_f_5:idx_f_1]

    def Get_Analytic_Transfer_Function(self):
        #Response function approximation from Calculation described by Cornish, Robson, Liu 2019
        if isinstance(self.f_low,u.Quantity) and isinstance(self.f_low,u.Quantity):
            self.fT = np.logspace(np.log10(self.f_low.value),np.log10(self.f_high.value),self.nfreqs)*u.Hz
        else:
            self.fT = np.logspace(np.log10(self.f_low),np.log10(self.f_high),self.nfreqs)*u.Hz
        f_L = const.c/2/np.pi/self.L #Transfer frequency
        #3/10 is normalization 2/5sin(openingangle)
        R_f = 3/10/(1+0.6*(self.fT/f_L)**2)
        self.transferfunction = np.sqrt(R_f)

    def Set_T_Function_Type(self,calc_type):
        if calc_type == 'n' or calc_type == 'N':
            self._T_Function_Type = 'numeric'
        elif calc_type == 'a' or calc_type == 'A':
            self._T_Function_Type = 'analytic'
        else:
            print('\nYou can get the transfer function via 2 methods:')
            print(' *To use the numerically approximated method in Robson, Cornish, and Liu, 2019, input "N".')
            print(' *To use the analytic fit in Larson, Hiscock, and Hellings, 2000, input "A".')
            calc_type = input('Please select the calculation type: ')
            self.Set_T_Function_Type(calc_type)
        if hasattr(self,'_T_Function_Type'):
            if self._T_Function_Type == 'numeric':
                self.Get_Numeric_Transfer_Function()
            if self._T_Function_Type == 'analytic':
                self.Get_Analytic_Transfer_Function()


    def Add_Background(self):
        """
        Galactic confusions noise parameters for 6months, 1yr, 2yr, and 4yr
        corresponding to array index 0,1,2,3 respectively
        """
        A = 9e-45
        a = np.array([0.133,0.171,0.165,0.138])
        b = np.array([243,292,299,-221])
        k = np.array([482,1020,611,521])
        g = np.array([917,1680,1340,1680])
        f_k = np.array([0.00258,0.00215,0.00173,0.00113])

        if self.T_obs < 1.*u.yr:
            index = 0
        elif self.T_obs >= 1.*u.yr and self.T_obs < 2.*u.yr:
            index = 1
        elif self.T_obs >= 2.*u.yr and self.T_obs < 4.*u.yr:
            index = 2
        else:
            index = 3
        f = self.fT.value
        return A*np.exp(-(f**a[index])+(b[index]*f*np.sin(k[index]*f)))\
                *(f**(-7/3))*(1 + np.tanh(g[index]*(f_k[index]-f))) #White Dwarf Background Noise



import numpy as np
import os
import astropy.constants as const
import astropy.units as u
import scipy.interpolate as interp
from astropy.cosmology import z_at_value
from astropy.cosmology import WMAP9 as cosmo

from .waveform import Get_Waveform
from . import utils

current_path = os.getcwd()
splt_path = current_path.split("/")
top_path_idx = splt_path.index('gwent')
top_directory = "/".join(splt_path[0:top_path_idx+1])

class BinaryBlackHole:
    """Base Class for frequency domain strains from Binary Black Holes.

    Parameters
    ----------
    M : float
        Total mass of the black hole binary (m1+m2)
    q : float
        Mass ratio of the black hole binary (m1/m2, m1<m2)
    z : float
        Redshift of the black hole binary

    load_location : string, optional
        the directory of the loaded file, (ie. '/path/to/file')

    Notes
    -----
    IMRPhenomD waveforms calibrated for q = m1/m2 < 18

    """
    def __init__(self,*args,**kwargs):
        if len(args) == 3:
            [M,q,z] = args
        elif len(args) == 6:
            [M,q,_,_,z,_] = args
        else:
            raise ValueError('args must be a list of 3 ([M,q,z]) or 6 ([M,q,z,chi1,chi2,inc')
        self.M = M
        self.q = q
        self.z = z

        for keys,value in kwargs.items():
            if keys == 'load_location':
                self.load_location = value

        if hasattr(self,'load_location'):
            self.Load_Data()

    @property
    def M(self):
        self._M = utils.make_quant(self._M,'M_sun')
        return self._M
    @M.setter
    def M(self,value):
        self.var_dict = ['M',value]
        self._M = self._return_value

    @property
    def q(self):
        return self._q
    @q.setter
    def q(self,value):
        self.var_dict = ['q',value]
        self._q = self._return_value

    @property
    def z(self):
        return self._z
    @z.setter
    def z(self,value):
        self.var_dict = ['z',value]
        self._z = self._return_value

    @property
    def h_f(self):
        if not hasattr(self,'_h_f'):
            raise NotImplementedError('The strain must be defined inside BBHFrequencyDomain or BBHTimeDomain classes.')
        return self._h_f
    @h_f.setter
    def h_f(self,value):
        self._h_f = value

    @property
    def f(self):
        if not hasattr(self,'_f'):
            raise NotImplementedError('Interferometer frequency must be defined inside SpaceBased or GroundBased classes.')
        return self._f
    @f.setter
    def f(self,value):
        self._f = value

    @property
    def var_dict(self):
        return self._var_dict
    @var_dict.setter
    def var_dict(self,value):
        utils.Get_Var_Dict(self,value)

    def Load_Data(self):
        if hasattr(self,'load_location'):
            if os.path.exists(self.load_location):
                self._load_data = np.loadtxt(self.load_location)
            else:
                raise IOError('File %s does not exist, please assign load_location a correct filepath.' %self.load_location)
        else:
            raise ValueError('load_location is not assigned, please set with name_of_BBH.load_location="path/to/file".')

class BBHFrequencyDomain(BinaryBlackHole):
    """Subclass of BinaryBlackHole for BBH GWs generated in the frequency domain.

    Parameters
    ----------
    chi1 : float
        The dimensionless spin parameter abs(a/m) for black hole m1.
    chi2 : float
        The dimensionless spin parameter abs(a/m) for black hole m2
    inc : float
        The inclination of the BBH

    f_low : float, optional
        The lowest frequency in natural units (Mf, G=c=1) at which the BBH waveform is calculated
    nfreqs : int, optional
        The number of frequencies at which the BBH waveform is calculated

    Notes
    -----
    IMRPhenomD waveforms calibrated for aligned spins chi_1, chi_2 = abs(a/m) <= 0.85 or if q=1 abs(a/m)<0.98

    """
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        [_,_,_,chi1,chi2,inc] = args
        self.chi1 = chi1
        self.chi2 = chi2
        self.inc = inc

        for keys,value in kwargs.items():
            if keys == 'f_low':
                self.f_low = value
            elif keys == 'f_high':
                self.f_high = value
            elif keys == 'nfreqs':
                self.nfreqs = value
            elif keys == 'instrument':
                self.instrument = value
                self.Check_Freq_Evol()
        if not hasattr(self,'nfreqs'):
            self.nfreqs = int(1e3)
        if not hasattr(self,'f_low'):
            self.f_low = 1e-5*u.Hz

        self.Get_Fitcoeffs()

    @property
    def chi1(self):
        return self._chi1
    @chi1.setter
    def chi1(self,value):
        self.var_dict = ['chi1',value]
        self._chi1 = self._return_value

    @property
    def chi2(self):
        return self._chi2
    @chi2.setter
    def chi2(self,value):
        self.var_dict = ['chi2',value]
        self._chi2 = self._return_value

    @property
    def inc(self):
        return self._inc
    @inc.setter
    def inc(self,value):
        self.var_dict = ['inc',value]
        self._inc = self._return_value

    @property
    def instrument(self):
        return self._instrument
    @instrument.setter
    def instrument(self,value):
        self._instrument = value

    @property
    def h_gw(self):
        if not hasattr(self,'_h_gw'):
            if not hasattr(self,'f_init'):
                if hasattr(self,'_instrument'):
                    self.Check_Freq_Evol()
                else:
                    raise ValueError('No instrument assigned, please fix it. '\
                        'Try: "source.instrument = instrument".')
                self._h_gw = Get_Mono_Strain(self,self.instrument.f_opt).to('')
            else:
                self._h_gw = Get_Mono_Strain(self,self.f_init).to('')
        return self._h_gw
    @h_gw.setter
    def h_gw(self,value):
        self._h_gw = value
    @h_gw.deleter
    def h_gw(self):
        del self._h_gw

    @property
    def h_f(self):
        if not hasattr(self,'_h_f'):
            if not (hasattr(self,'_phenomD_f') and hasattr(self,'_phenomD_h')):
                self.Get_PhenomD_Strain()
            [_,self._h_f] = Strain_Conv(self,self._phenomD_f,self._phenomD_h)
        return self._h_f
    @h_f.deleter
    def h_f(self):
        del self._h_f

    @property
    def f(self):
        if not hasattr(self,'_f'):
            if not (hasattr(self,'_phenomD_f') and hasattr(self,'_phenomD_h')):
                self.Get_PhenomD_Strain()
            [self._f,_] = Strain_Conv(self,self._phenomD_f,self._phenomD_h)
        return self._f
    @f.deleter
    def f(self):
        del self._f

    def Get_Fitcoeffs(self):
        """Loads Quasi-Normal Mode fitting files for speed later."""
        fit_coeffs_filedirectory = top_directory + '/docs/LoadFiles/PhenomDFiles/fitcoeffsWEB.dat'
        self._fitcoeffs = np.loadtxt(fit_coeffs_filedirectory)

    def Get_PhenomD_Strain(self):
        """Gets the BBH's frequency and waveform from IMRPhenomD."""
        if not hasattr(self,'_fitcoeffs'):
            self.Get_Fitcoeffs()
        [self._phenomD_f,self._phenomD_h] = Get_Waveform(self)

    def Get_Time_From_Merger(self,f_obs):
        """Calculates the time from merger of a binary black hole given an observed frequency.

        Parameters
        ----------
        f_obs : float
            the initially observed frequency in the instrument frame.

        """
        m_conv = const.G/const.c**3 #Converts M = [M] to M = [sec]
        eta = self.q/(1+self.q)**2

        M_time = self.M.to('kg')*m_conv
        M_chirp = eta**(3/5)*M_time

        f_obs_source = f_obs*(1+self.z)
        return 5*(M_chirp)**(-5/3)*(8*np.pi*f_obs_source)**(-8/3)

    def Get_Source_Freq(self,tau):
        """Calculates the binary black hole's gravitational wave frequency given a time from merger

        Parameters
        ----------
        tau : float
            the time from merger in the source frame

        """
        m_conv = const.G/const.c**3 #Converts M = [M] to M = [sec]
        eta = self.q/(1+self.q)**2

        M_time = self.M.to('kg')*m_conv
        M_chirp = eta**(3/5)*M_time

        return 1./8./np.pi/M_chirp*(5*M_chirp/tau)**(3./8.)

    def Check_Freq_Evol(self):
        """Checks the frequency evolution of the black hole binary.

        Notes
        -----
        If the frequency of the binary does evolve over more than one bin,
        (ie f(T_obs)-f(t_init) = delf_obs < 1/T_obs), it is monochromatic, so we set the frequency
        to the optimal frequency of the detector

        Otherwise it is chirping and evolves over the observation and we
        set the starting frequency we observe it at to f(Tobs), which is the
        frequency at an observation time before merger

        To get the change in frequency, we use eqn 41 from Hazboun,Romano, and Smith (2019) <https://arxiv.org/abs/1907.04341>
        which uses binomial expansion of f_T_obs_inst - f_init_inst and thus will never be imaginary

        """
        m_conv = const.G/const.c**3 #Converts M = [M] to M = [sec]
        eta = self.q/(1+self.q)**2

        M_time = self.M.to('kg')*m_conv
        M_chirp_source = eta**(3/5)*M_time

        T_obs = utils.make_quant(self.instrument.T_obs,'s')
        T_obs_source = T_obs/(1+self.z)


        #Assumes t_init is in source frame, can either be randomly drawn
        #t_init_source = np.random.uniform(0,100)*u.yr

        #Assumes f_init is the optimal frequency in the instrument frame to get t_init_source
        self.f_init = self.instrument.f_opt
        t_init_source = self.Get_Time_From_Merger(self.f_init)

        #f(T_obs), the frequency of the source at T_obs before merger
        f_T_obs_source = self.Get_Source_Freq(T_obs_source)
        #f(T_obs) in the instrument frame
        self.f_T_obs = f_T_obs_source/(1+self.z)

        #t_init_source = make_quant(t_init_source,'s')
        #f_init_source = self.Get_Source_Freq(t_init_source)
        #self.f_init = f_init_source/(1+self.z)
        #f_after_T_obs_source = self.Get_Source_Freq((t_init_source-T_obs_source))
        #self.f_T_obs = f_after_T_obs_source/(1+self.z)
        #delf_obs_source_exact = f_after_T_obs_source-f_init_source

        delf_obs_source_approx = 1./8./np.pi/M_chirp_source*(5*M_chirp_source/t_init_source)**(3./8.)*(3*T_obs_source/8/t_init_source)
        delf_obs =  delf_obs_source_approx/(1+self.z)

        if delf_obs < (1/T_obs):
            self.ismono = True
        else:
            self.ismono = False




class BBHTimeDomain(BinaryBlackHole):
    """Subclass of BinaryBlackHole for input in the time domain"""
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.Get_hf_from_hcross_hplus()

    @property
    def t(self):
        if not hasattr(self,'_t'):
            self._t = self._load_data[:,0]
        self._t = utils.make_quant(self._t,'s')
        return self._t

    @property
    def h_plus_t(self):
        if not hasattr(self,'_h_plus_t'):
            self._h_plus_t = self._load_data[:,1]
        return self._h_plus_t

    @property
    def h_cross_t(self):
        if not hasattr(self,'_h_cross_t'):
            self._h_cross_t = self._load_data[:,1]
        return self._h_cross_t

    @property
    def h_f(self):
        if not hasattr(self,'_h_f'):
            [natural_f,natural_h] = self.Get_hf_from_hcross_hplus()
            [_,self._h_f] = Strain_Conv(self,natural_f,natural_h)
        return self._h_f
    @h_f.deleter
    def h_f(self):
        del self._h_f

    @property
    def f(self):
        if not hasattr(self,'_f'):
            [natural_f,natural_h] = self.Get_hf_from_hcross_hplus()
            [self._f,_] = Strain_Conv(self,natural_f,natural_h)
        return self._f
    @f.deleter
    def f(self):
        del self._f


    def Get_hf_from_hcross_hplus(self,interp_res='coarse',windowing='left'):
        """Converts dimensionless, time domain strain to frequency space using a windowed fft

        Parameters
        ----------
        interp_res : {'coarse','fine'}, optional
            'coarse' uses maximum difference between subsequent time steps for interpolation
            'fine' uses minimum difference between subsequent time steps for interpolation
        windowing : {'left','right','all'}, optional
            'left' windows the left side of the time data
            'right' windows the right side of the time data
            'all' windows the both the left and right side of the time data

        Returns
        -------
        natural_f : array
            The frequency of the input source in natural units (G=c=1)
        natural_h : array
            The strain of the input source in natural units (G=c=1)

        """

        #Interpolate time to evenly sampled data, can be fine or coarse
        diff_t = np.diff(self.t.value)
        if interp_res == 'fine':
            dt = min(diff_t)
        elif interp_res == 'coarse':
            dt = max(diff_t)

        interp_t = np.arange(self.t[0].value,self.t[-1].value,dt)
        #interpolate strain to evenly sampled data for FFT
        h_cross_t = interp.interp1d(self.t,self.h_cross_t,kind='cubic')
        h_plus_t = interp.interp1d(self.t,self.h_plus_t,kind='cubic')
        interp_h_cross_t = h_cross_t(interp_t)
        interp_h_plus_t = h_plus_t(interp_t)

        #Filter/Window
        hann_window = np.hanning(len(interp_t)) #Two sided
        if windowing == 'left':
            #########################
            """Applies window to first (left) half"""
            first_half = hann_window[:int(len(interp_t)/2)] # Only need tapering on first half of waveform
            second_half = np.ones(len(interp_t)-len(first_half)) #no windowing on second half of waveform
            #########################
            window = np.append(first_half,second_half) # Only apply window to first half of waveform
        elif windowing == 'right':
            #########################
            """Applies window to second (right) half"""
            second_half = hann_window[int(len(interp_t)/2):] # Only need tapering on second half of waveform
            first_half = np.ones(len(interp_t)-len(second_half)) #no windowing on first half of waveform
            #########################
            window = np.append(first_half,second_half)
        elif windowing == 'all':
            window = hann_window
        #Window!
        win_h_cross_t = np.multiply(interp_h_cross_t,window)
        win_h_plus_t = np.multiply(interp_h_plus_t,window)

        #FFT the two polarizations
        h_cross_f = np.fft.fft(win_h_cross_t)
        h_plus_f = np.fft.fft(win_h_plus_t)
        freqs = np.fft.fftfreq(len(interp_t),d=dt)

        #cut = np.abs(freqs).argmax() #Cut off the negative frequencies
        f_cut_low = 3e-3 #Low Cutoff frequency
        f_cut_high = 1.5e-1 #High Cutoff frequency
        cut_low = np.abs(freqs-f_cut_low).argmin() #Cut off frequencies lower than a frequency
        cut_high = np.abs(freqs-f_cut_high).argmin() #Cut off frequencies higher than a frequency
        #cut=int(len(freqs)*0.9) #Cut off percentage of frequencies
        h_cross_f = h_cross_f[cut_low:cut_high]
        h_plus_f = h_plus_f[cut_low:cut_high]
        natural_f = freqs[cut_low:cut_high]

        #Combine them for raw spectral power
        natural_h_f = np.sqrt((np.abs(h_cross_f))**2 + (np.abs(h_plus_f))**2)
        return [natural_f,natural_h_f]



def Strain_Conv(source,natural_f,natural_h):
    """Converts frequency and strain in natural units (G=c=1) to Hertz and dimensionless, respectively.

    Parameters
    ----------
    source
        Instance of gravitational wave source class
    natural_f : array [Mf]
        the frequency of the source in natural units (G=c=1)
    natural_h : array [Mf]
        the strain of the source in natural units (G=c=1)

    """
    DL = cosmo.luminosity_distance(source.z)
    DL = DL.to('m')

    m_conv = const.G/const.c**3 #Converts M = [M] to M = [sec]
    M_redshifted_time = source.M.to('kg')*(1+source.z)*m_conv

    #frequency and strain of source in detector frame
    freq_conv = 1/M_redshifted_time
    #Normalized factor to match Stationary phase approx at low frequencies?
    #Changed from sqrt(5/16/pi)
    strain_conv = np.sqrt(1/4/np.pi)*(const.c/DL)*M_redshifted_time**2

    f = natural_f*freq_conv
    h_f = natural_h*strain_conv
    return [f,h_f]

def Get_Char_Strain(source):
    """Converts source strain to characteristic strain

    Parameters
    ----------
    source
        Instance of gravitational wave source class

    """
    h_char = np.sqrt(4*source.f**2*source.h_f**2)
    return h_char

def Get_Mono_Strain(source,f_gw,strain_const='Averaged'):
    """Calculates the strain from a binary black hole.

    Parameters
    ----------
    f_gw : float
        The source frequency of the gravitational wave.
    strain_const : {'Averaged','UseInc','Optimal'}, optional
        'Averaged' gives the sky and inclination averaged strain from Robson et al. 2019 (eqn 27) <https://arxiv.org/pdf/1803.01944.pdf>
        'UseInc' uses the source inclination value from Rosado, Sesana, and Gair (2015) <https://arxiv.org/abs/1503.04803>, will be innacurate because nothing else accounts for inclination
        'Optimal' gives the optimally oriented, face-on, inclination (ie. inc=0) value

    Returns
    -------
    float
        the strain of a monochromatic source in the dector frame

    """
    f_gw = utils.make_quant(f_gw,'Hz')
    if isinstance(strain_const,str):
        DL = cosmo.luminosity_distance(source.z)
        DL = DL.to('m')

        #Converts M = [M] to M = [sec]
        m_conv = const.G/const.c**3

        eta = source.q/(1+source.q)**2
        M_redshifted_time = source.M.to('kg')*(1+source.z)*m_conv
        M_chirp = eta**(3/5)*M_redshifted_time

        if strain_const == 'UseInc':
            a = 1+np.cos(source.inc)**2
            b = -2*np.cos(source.inc)
            const_val = 2*np.sqrt(.5*(a**2+b**2))
        elif strain_const == 'Optimal':
            const_val = 4.
        elif strain_const == 'Averaged':
            const_val = 8/np.sqrt(5)
        else:
            raise ValueError('Can only use "UseInc", "Averaged", or "Optimal" monochromatic strain calculation.')

        return const_val*(const.c/DL)*(np.pi*f_gw)**(2./3.)*M_chirp**(5./3.)
    else:
        raise ValueError('Can only use "UseInc", "Averaged", or "Optimal" monochromatic strain calculation.')



import numpy as np

def Get_Waveform(source,pct_of_peak=0.01):
    """Uses Mass Ratio (q <= 18), aligned spins (abs(a/m)~0.85 or when q=1 abs(a/m)<0.98),
    fitting coefficients for QNM type, and sampling rate
    Returns the frequency, the Phenom amplitude of the inspiral-merger-ringdown
    Uses methods found in <https://arxiv.org/abs/1508.07253> and <https://arxiv.org/abs/1508.07250>

    Parameters
    ----------
    source : object
            source object from StrainandNoise, contains all source parameters
    pct_of_peak : float, optional
            the percentange of the strain at merger that dictates the maximum frequency the waveform is calculated at in geometrized units (G=c=1)

    Returns
    -------
    Mf : numpy array of floats
        the waveform frequencies in geometrized units (G=c=1)
    fullwaveform : numpy array of floats
        the waveform strain in geometrized units (G=c=1)

    """
    f_low = source.f_low.value
    N = source.nfreqs
    q = source.q
    x1 = source.chi1
    x2 = source.chi2
    fitcoeffs = source._fitcoeffs

    #M = m1+m2 #Total Mass
    #q = m2/m1 #Mass Ratio: Paper tested up to 18
    #eta = m1*m2/M**2 reduced mass: Paper tested up to 0.05 (q=18)
    eta = q/(q+1)**2
    x_PN = chi_PN(eta,x1,x2) #PN reduced spin parameter
    a_f = a_final(x1,x2,q,eta) #dimensionless spin

    ##################
    #Finds f_ringdown and f_damp from fit taken from <https://arxiv.org/abs/gr-qc/0512160>
    n = 0      #QNM indices
    l = 2
    m = 2
    numn = 3   #number of n's included in the table

    index = (l-2)*(2*l+1)*numn + (l-m)*numn + n
    f_fit = fitcoeffs[index][3:6]
    q_fit = fitcoeffs[index][6:9]

    omega_RD = f_fit[0]+f_fit[1]*(1-a_f)**f_fit[2]          #M omega_{lmn}
    tau = 2*(q_fit[0]+q_fit[1]*(1-a_f)**q_fit[2])/omega_RD  #tau_{lmn}/M = 2 Q_{lmn}/(M omega_{lmn})
    ########################
    f_RD = omega_RD/2/np.pi
    f_damp = 1/tau/2/np.pi

    Gamma1 = Lambda(eta,x_PN,4)
    Gamma2 = Lambda(eta,x_PN,5)
    Gamma3 = Lambda(eta,x_PN,6)

    f_peak = Calc_f_peak(f_RD,f_damp,[Gamma1,Gamma2,Gamma3])

    f1 = 0.014
    f3 = f_peak
    f2 = (f1+f3)/2

    cutoffFreq = Find_Cutoff_Freq(f_RD,f_damp,[Gamma1,Gamma2,Gamma3],pct_of_peak=pct_of_peak)

    #If lowest frequency is lower than cutoffFreq, throw error
    if f_low <= cutoffFreq:
        raise ValueError('Lower frequency bound must be lower than that of the merger ringdown.')
    
    Mf = np.logspace(np.log10(f_low),np.log10(cutoffFreq),N)

    v1 = A_insp(f1,eta,x1,x2,x_PN)
    v2 = Lambda(eta,x_PN,3)
    v3 = A_MR(f3,f_RD,f_damp,[Gamma1,Gamma2,Gamma3])
    fund1 = DA_insp(f1,eta,x1,x2,x_PN)
    fund3 = DA_MR(f3,f_RD,f_damp,[Gamma1,Gamma2,Gamma3])

    #############################
    #Calculate Solutions to eqn 21 in intermediate region
    Del_solns = A_intermediate(f1,f2,f3,v1,v2,v3,fund1,fund3) # Solutions to eqn 21

    ##############################
    #Calculate all sections of waveform and Paste together
    indxf1 = np.argmin(np.abs(Mf-f1))
    indxfpeak = np.argmin(np.abs(Mf-f_peak))

    tmpinspiral = A_norm(Mf[0:indxf1+1],eta)*A_insp(Mf[0:indxf1+1],eta,x1,x2,x_PN)
    tmpintermediate = A_norm(Mf[indxf1+1:indxfpeak],eta)*A_int(Mf[indxf1+1:indxfpeak],Del_solns)
    tmpmergerringdown = A_norm(Mf[indxfpeak:],eta)*A_MR(Mf[indxfpeak:],f_RD,f_damp,[Gamma1,Gamma2,Gamma3])
    fullwaveform = np.hstack((tmpinspiral,tmpintermediate,tmpmergerringdown))

    return [Mf,fullwaveform]

def A_norm(freqs,eta):
    """Calculates the constant scaling factor A_0

    Parameters
    ----------
    freqs : array
        The frequencies in Natural units (Mf, G=c=1) of the waveform
    eta : float
        The reduced mass ratio

    """
    const = np.sqrt(2*eta/3/np.pi**(1/3))
    return const*freqs**-(7/6)


def A_insp(freqs,eta,x1,x2,X_PN):
    """Calculates the Inspiral Amplitude

    Parameters
    ----------
    freqs : array
        The frequencies in Natural units (Mf, G=c=1) of the waveform
    eta : float
        The reduced mass ratio
    x1 : float
        The dimensionless spin parameter abs(a/m) for black hole m1.
    x2 : float
        The dimensionless spin parameter abs(a/m) for black hole m2.
    x_PN : float
        The PN reduced spin parameter

    """
    A_PN = 0.0
    A_higher = 0.0
    for i in range(7):
        A_PN = A_PN + PN_coeffs(eta,x1,x2,i)*(np.pi*freqs)**(i/3)
        if i >= 1 and i <= 3:
            A_higher = A_higher + Lambda(eta,X_PN,i-1)*freqs**((6+i)/3)
    return (A_PN + A_higher)


def DA_insp(freqs,eta,x1,x2,X_PN):
    """Calculates Derivative of the inspiral amplitude.

    Parameters
    ----------
    freqs : array
        The frequencies in Natural units (Mf, G=c=1) of the waveform
    eta : float
        The reduced mass ratio
    x1 : float
        The dimensionless spin parameter abs(a/m) for black hole m1.
    x2 : float
        The dimensionless spin parameter abs(a/m) for black hole m2.
    x_PN : float
        The PN reduced spin parameter

    """
    DA_PN = 0.0
    DA_higher = 0.0
    for i in range(7):
        PN_const = np.pi**(i/3)*(i/3)*PN_coeffs(eta,x1,x2,i)
        DA_PN = DA_PN + PN_const*(freqs)**((i-3)/3)
        if i >= 1 and i <= 3:
            higher_const = ((6+i)/3)*Lambda(eta,X_PN,i-1)
            DA_higher = DA_higher + higher_const*freqs**((i+3)/3)
        
    return DA_PN + DA_higher


def A_MR(freqs,f_RD,f_damp,Gammas):
    """Calculates the Normalized Merger-Ringdown Amplitude
    
    Parameters
    ----------
    freqs : array
        The frequencies in Natural units (Mf, G=c=1) of the waveform
    f_RD : float
        Frequency of the Ringdown transition
    f_damp : float
        Damping frequency
    Gammas : array-like
        Normalizes lorentzian to correct shape

    """
    varf = freqs-f_RD
    fg_d = Gammas[2]*f_damp 
    return (Gammas[0]*fg_d)/(varf**2+fg_d**2)*np.exp(-(Gammas[1]/fg_d)*varf)


def DA_MR(freqs,f_RD,f_damp,Gammas):
    """Calculates Derivative of the Merger-Ringdown Amplitude
    
    Parameters
    ----------
    freqs : array
        The frequencies in Natural units (Mf, G=c=1) of the waveform
    f_RD : float
        Frequency of the Ringdown transition
    f_damp : float
        Damping frequency
    Gammas : array-like
        Normalizes lorentzian to correct shape

    """
    varf = freqs-f_RD
    fg_d = Gammas[2]*f_damp
    A_MR_0 = A_MR(freqs,f_RD,f_damp,Gammas)
    return -A_MR_0*(2*varf/(varf**2+fg_d**2)+Gammas[1]/fg_d)


def A_intermediate(f1,f2,f3,v1,v2,v3,d1,d3):
    """Solves system of equations for intermediate amplitude matching"""
    Mat = np.array([[1., f1, f1**2, f1**3, f1**4],[1., f2, f2**2, f2**3, f2**4],[1., f3, f3**2, f3**3, f3**4], \
            [0., 1., 2*f1, 3*f1**2, 4*f1**3],[0., 1., 2*f3, 3*f3**2, 4*f3**3]],dtype='float')
    a = np.array([v1,v2,v3,d1,d3],dtype='float')
    return np.linalg.solve(Mat,a)


def A_int(freqs,delt):
    """Calculates the Intermediate Amplitude

    Parameters
    ----------
    freqs : array
        The frequencies in Natural units (Mf, G=c=1) of the waveform
    delt : array
        Coefficient solutions to match the inspiral to the merger-ringdown portion of the waveform

    """
    return (delt[0]+delt[1]*freqs+delt[2]*freqs**2+delt[3]*freqs**3+delt[4]*freqs**4)


def Lambda(eta,x_PN,lmbda):
    """Gets the Lambdas from Eqn 31 in <https://arxiv.org/abs/1508.07253>

    Parameters
    ----------
    eta : float
        The reduced mass ratio
    x_PN : float
        The PN reduced spin parameter
    lmbda : int
        Iterator for different Lambda variables using the zeta function

    """
    xi = x_PN-1
    xi2 = xi*xi
    xi3 = xi2*xi
    eta2 = eta*eta
    if lmbda == 0:  #rho1
        coeffs = zeta(0)
    elif lmbda == 1: #rho2
        coeffs = zeta(1)
    elif lmbda == 2: #rho3
        coeffs = zeta(2)
    elif lmbda == 3: #v2
        coeffs = zeta(3)
    elif lmbda == 4: #gamma1
        coeffs = zeta(4)
    elif lmbda == 5: #gamma2
        coeffs = zeta(5)
    elif lmbda == 6: #gamma3
        coeffs = zeta(6)
    
    return coeffs[0] + coeffs[1]*eta + \
        (coeffs[2] + coeffs[3]*eta + coeffs[4]*eta2)*xi + \
        (coeffs[5] + coeffs[6]*eta + coeffs[7]*eta2)*xi2 + \
        (coeffs[8] + coeffs[9]*eta + coeffs[10]*eta2)*xi3


def zeta(k):
    """Coefficients in table 5 of <https://arxiv.org/abs/1508.07253>"""
    if k == 0: #rho 1
        coeffs = [3931.9, -17395.8, 3132.38, 343966.0, -1.21626e6, -70698.0, 1.38391e6, -3.96628e6, -60017.5, 803515.0, -2.09171e6]
    elif k == 1: #rho 2
        coeffs = [-40105.5, 112253.0, 23561.7, -3.47618e6, 1.13759e7, 754313.0, -1.30848e7, 3.64446e7, 596227.0, -7.42779e6, 1.8929e7]
    elif k == 2: #rho 3
        coeffs = [83208.4, -191238.0, -210916.0, 8.71798e6, -2.69149e7, -1.98898e6, 3.0888e7, -8.39087e7, -1.4535e6, 1.70635e7, -4.27487e7]
    elif k == 3: #v 2
        coeffs = [0.814984, 2.57476, 1.16102, -2.36278, 6.77104, 0.757078, -2.72569, 7.11404, 0.176693, -0.797869, 2.11624]
    elif k == 4: #gamma 1
        coeffs = [0.0069274, 0.0302047, 0.00630802, -0.120741, 0.262716, 0.00341518, -0.107793, 0.27099, 0.000737419, -0.0274962, 0.0733151]
    elif k == 5: #gamma 2
        coeffs = [1.01034, 0.000899312, 0.283949, -4.04975, 13.2078, 0.103963, -7.02506, 24.7849, 0.030932, -2.6924, 9.60937]
    elif k == 6: #gamma 3
        coeffs = [1.30816, -0.00553773, -0.0678292, -0.668983, 3.40315, -0.0529658, -0.992379, 4.82068, -0.00613414, -0.384293, 1.75618]
    return coeffs


def PN_coeffs(eta,x1,x2,i):
    """Gets the PN Amplitude coefficients

    Parameters
    ----------
    eta : float
        The reduced mass ratio
    x1 : float
        The dimensionless spin parameter abs(a/m) for black hole m1.
    x2 : float
        The dimensionless spin parameter abs(a/m) for black hole m2.
    q : float
        The mass ratio m1/m2, m1<=m2
    i : int
        iterator to dictate which PN Amplitude to use

    Notes
    -----
    Coefficients in appendix B (eqns B14-B20) of <https://arxiv.org/abs/1508.07253>

    """
    delta = np.sqrt(1.0-4.0*eta)
    chi_s = (x1+x2)/2.0
    chi_a = (x1-x2)/2.0
    if i == 0:
        A_i = 1
    elif i == 1:
        A_i = 0
    elif i == 2:
        A_i = -323/224 + (451/168)*eta
    elif i == 3:
        A_i = (27/8)*delta*chi_a + (27/8 -(11/6)*eta)*chi_s
    elif i == 4:
        A_i = -27312085/8128512 -(1975055/338688)*eta + (105271/24192)*eta**2 + \
            (-81/32+8*eta)*chi_a**2 - 81/16*delta*chi_a*chi_s + (-81/32+17/8*eta)*chi_s**2
    elif i == 5:
        A_i = -85*np.pi/64 + 85*np.pi/16*eta + (285197/16128-1579/4032*eta)*delta*chi_a + \
            (285197/16128 - 15317/672*eta - 2227/1008*eta**2)*chi_s
    elif i == 6:
        A_i = -177520268561/8583708672 + (545384828789/5007163392 - 205*np.pi**2/48)*eta - \
            3248849057/178827264*eta**2 + 34473079/6386688*eta**3 + (1614569/64512-1873643/16128*eta+2167/42*eta**2)*chi_a**2 + \
            (31*np.pi/12 - 7*np.pi/3*eta)*chi_s + (1614569/64512-61391/1344*eta+57451/4032*eta**2)*chi_s**2 + \
            delta*chi_a*(31*np.pi/12+(1614569/32256-165961/2688*eta)*chi_s)
    return A_i


def Calc_f_peak(f_RD,f_damp,Gammas):
    """Calculates the frequency at the peak of the merger

    Parameters
    ----------
    f_RD : float
        Frequency of the Ringdown transition
    f_damp : float
        Damping frequency
    Gammas : array-like
        Normalizes lorentzian to correct shape
    
    Notes
    -----
    There is a problem with this expression from the paper becoming imaginary if gamma2 >= 1 
    so if gamma2 >= 1 then set the square root term to zero.

    """
    if Gammas[1] <= 1:
        f_max = np.abs(f_RD+f_damp*Gammas[2]*(np.sqrt(1-Gammas[1]**2)-1)/Gammas[1])
    else:
        f_max = np.abs(f_RD+(f_damp*Gammas[2]*-1)/Gammas[1])
    return f_max


def Find_Cutoff_Freq(f_RD,f_damp,Gammas,pct_of_peak=0.0001):
    """Cutoff signal when the amplitude is a factor of 10 below the value at f_RD

    Parameters
    ----------
    f_RD : float
        Frequency of the Ringdown transition
    f_damp : float
        Damping frequency
    Gammas : array-like
        Normalizes lorentzian to correct shape

    pct_of_peak : float, optional
        the percentange of the strain at merger that dictates the maximum 
        frequency the waveform is calculated at in geometrized units (G=c=1) 

    """
    tempfreqs = np.logspace(np.log10(f_RD),np.log10(10*f_RD),100)
    cutoffAmp = pct_of_peak*A_MR(f_RD,f_RD,f_damp,[Gammas[0],Gammas[1],Gammas[2]])
    merger_ringdown_Amp = A_MR(tempfreqs,f_RD,f_damp,[Gammas[0],Gammas[1],Gammas[2]])
    cutoffindex = np.argmin(np.abs(cutoffAmp-merger_ringdown_Amp))
    return tempfreqs[cutoffindex]

def a_final(x1,x2,q,eta):
    """The Final spin of the binary remnant black hole

    Parameters
    ----------
    x1 : float
        The dimensionless spin parameter abs(a/m) for black hole m1.
    x2 : float
        The dimensionless spin parameter abs(a/m) for black hole m2.
    q : float
        The mass ratio m1/m2, m1<=m2
    eta : float
        The reduced mass ratio

    Notes
    -----
    Uses eq. 3 in <https://arxiv.org/abs/0904.2577>, changed to match our q convention
    a=J/M**2 where J = x1*m1**2 + x2*m2**2

    """
    a = (q**2*x1+x2)/(q**2+1)
    s4 = -0.1229
    s5 = 0.4537
    t0 = -2.8904
    t2 = -3.5171
    t3 = 2.5763
    return a + s4*a**2*eta + s5*a*eta**2 + t0*a*eta + 2*np.sqrt(3)*eta + t2*eta**2 + t3*eta**3


def chi_PN(eta,x1,x2):
    """Calculates the PN reduced spin parameter

    Parameters
    ----------
    eta : float
        The reduced mass ratio
    x1 : float
        The dimensionless spin parameter abs(a/m) for black hole m1.
    x2 : float
        The dimensionless spin parameter abs(a/m) for black hole m2.

    Notes
    -----
    See Eq 5.9 in <https://arxiv.org/abs/1107.1267v2>

    """
    delta = np.sqrt(1.0-4.0*eta)
    chi_s = (x1+x2)/2.0
    chi_a = (x1-x2)/2.0
    return chi_s*(1.0-eta*76.0/113.0) + delta*chi_a


import astropy.units as u

def make_quant(param, default_unit):
    """Convenience function to intialize a parameter as an astropy quantity.

    Parameters
    ----------
    param : float, or Astropy Quantity
        Parameter to initialize
    default_unit : str
        Astropy unit string, sets as default for param.

    Returns
    -------
        an astropy quantity
    
    Examples
    -------
        self.f0 = make_quant(f0,'MHz')

    Notes
    -----
    Taken from <https://github.com/Hazboun6/hasasia/blob/master/hasasia/sensitivity.py#L834>

    """
    default_unit = u.core.Unit(default_unit)
    if hasattr(param, 'unit'):
        try:
            quantity = param.to(default_unit)
        except u.UnitConversionError:
            raise ValueError("Quantity {0} with incompatible unit {1}"
                             .format(param, default_unit))
    else:
        quantity = param * default_unit

    return quantity


def Get_Var_Dict(obj,value):
    """Updates and initializes variable dictionaries used to keep track of
    current values and variable minima and maxima.

    Parameters
    ----------
    obj : object
        Instance of class with parameter variables
    value : array-like
        value(s) that are assigned into dictionary

    Notes
    -----
    value contains the variable name in the first index
    the next is the current value of the variable
    the last two are optional and contain the variable min and max

    Examples
    --------
    obj.var_dict = ['M',value]
        where obj is in this case an instance of a BinaryBlackHole

    """
    if not hasattr(obj,'var_dict'):
            obj._var_dict = {}
    if isinstance(value,list):
        if len(value) == 2 and isinstance(value[0],str):
            var_name = value[0]
            vals = value[1]
            if isinstance(vals,list) and len(vals) == 3:
                if isinstance(vals[0],(float,int,u.Quantity))\
                 and isinstance(vals[1],(float,int,u.Quantity))\
                  and isinstance(vals[2],(float,int,u.Quantity)):
                    obj._return_value = vals[0]
                    obj._var_dict[var_name] = {'val':vals[0],'min':vals[1],'max':vals[2]}
                else:
                    raise ValueError(DictError_3())
            elif isinstance(vals,(float,int,u.Quantity)):
                if isinstance(vals,(float,int,u.Quantity)):
                    if var_name in obj._var_dict.keys():
                        obj._var_dict[var_name]['val'] = vals
                    else:
                        obj.var_dict[var_name] = {'val':vals,'min':None,'max':None}
                    obj._return_value = vals
                else:
                    raise ValueError(DictError_2())
        else:
            raise ValueError(DictError_Full())
    else:
        raise ValueError(DictError_Full())

def DictError_Full():
    return 'Must assign either: \n\
    - A name and value in a list (ie. ["name",val]), or \n\
    - A name, a value, a minimum value, and maximum value in a list (ie. ["name",val,min,max]), \n\
    where where name is a string, and val,min,and max are either floats, ints, or an astropy Quantity.'
def DictError_3():
    return 'Must assign a name, a value, a minimum value, and maximum value in a list (ie. ["name",val,min,max]), \n\
    where name is a string, and val, min, and max are either floats, ints, or astropy Quantities.'
def DictError_2():
    return 'Must assign a name and value in a list (ie. ["name",val]) \n\
    where name is a string, and val is either a float, an int, or an astropy Quantity.'


import numpy as np

import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.image import NonUniformImage
import matplotlib.ticker as ticker
from matplotlib import cm

import astropy.constants as const
import astropy.units as u
from astropy.cosmology import z_at_value
from astropy.cosmology import WMAP9 as cosmo

def Plot_SNR(source,instrument,var_x,sample_x,var_y,sample_y,SNRMatrix,display=True,dl_axis=False,smooth_contours=False,figloc=None):
    """Plots the SNR contours from calcSNR

    Parameters
    ----------
    source : object
        Instance of a gravitational wave source class
    instrument : object
        Instance of a gravitational wave detector class
    var_x : str
        x-axis variable
    sample_x : array
        samples at which SNRMatrix was calculated corresponding to the x-axis variable
    var_y : str
        y-axis variable
    sample_y : array
        samples at which SNRMatrix was calculated corresponding to the y-axis variable
    SNRMatrix : array-like
        the matrix at which the SNR was calculated corresponding to the particular x and y-axis variable choices

    display : bool, optional
        Option to turn off display if saving multiple plots to a file
    dl_axis : bool, optional
        Option to turn on the right hand side labels of luminosity distance
    smooth_contours : bool, optional
        Option to interpolate contours to a finer mesh size to appear smooth instead of tiered contours
    figloc : str, optional
        If not None, saves the figure to the path figloc (ie. figloc = '/path/to/save/location/figname.type')

    """

    axissize = 16
    labelsize = 18
    textsize = 12
    textcolor = 'k'
    linesize = 4
    figsize = (10,8)

    colormap = 'viridis'
    logSNR = np.log10(SNRMatrix)

    logLevels_min = np.log10(np.array([5]))
    logLevels_max = np.ceil(np.amax(logSNR))
    print_logLevels = np.append(logLevels_min,np.arange(1,logLevels_max+1))
    if smooth_contours:
        logLevels = np.linspace(logLevels_min,logLevels_max,100)[:,0]
    else:
        logLevels = print_logLevels

    if var_x in source.var_dict.keys():
        if isinstance(source.var_dict[var_x]['min'],u.Quantity):
            xlabel_min = source.var_dict[var_x]['min'].value
            xlabel_max = source.var_dict[var_x]['max'].value
        else:
            xlabel_min = source.var_dict[var_x]['min']
            xlabel_max = source.var_dict[var_x]['max']
    elif var_x in instrument.var_dict.keys():
        if isinstance(instrument.var_dict[var_x]['min'],u.Quantity):
            xlabel_min = instrument.var_dict[var_x]['min'].value
            xlabel_max = instrument.var_dict[var_x]['max'].value
        else:
            xlabel_min = instrument.var_dict[var_x]['min']
            xlabel_max = instrument.var_dict[var_x]['max']
        if var_x == 'T_obs':
            xlabel_min = xlabel_min*u.yr.to('s')
            xlabel_max = xlabel_max*u.yr.to('s')
    else:
        raise ValueError(var_x + ' is not a variable in the source or the instrument.')

    if var_y in source.var_dict.keys():
        if isinstance(source.var_dict[var_y]['min'],u.Quantity):
            ylabel_min = source.var_dict[var_y]['min'].value
            ylabel_max = source.var_dict[var_y]['max'].value
        else:
            ylabel_min = source.var_dict[var_y]['min']
            ylabel_max = source.var_dict[var_y]['max']
    elif var_y in instrument.var_dict.keys():
        if isinstance(instrument.var_dict[var_y]['min'],u.Quantity):
            ylabel_min = instrument.var_dict[var_y]['min'].value
            ylabel_max = instrument.var_dict[var_y]['max'].value
        else:
            ylabel_min = instrument.var_dict[var_y]['min']
            ylabel_max = instrument.var_dict[var_y]['max']
        if var_y == 'T_obs':
            ylabel_min = ylabel_min*u.yr.to('s')
            ylabel_max = ylabel_max*u.yr.to('s')
    else:
        raise ValueError(var_y + ' is not a variable in the source or the instrument.')

    #Can't take log of astropy variables
    if isinstance(sample_x,u.Quantity):
        sample_x = sample_x.value
    if isinstance(sample_y,u.Quantity):
        sample_y = sample_y.value

    #Set whether log or linearly spaced axes
    if var_x == 'q' or var_x == 'chi1' or var_x == 'chi2':
        xaxis_type = 'lin'
    else:
        xaxis_type = 'log'

    if var_y == 'q' or var_y == 'chi1' or var_y == 'chi2':
        yaxis_type = 'lin'
    else:
        yaxis_type = 'log'

    #########################
    #Make the Contour Plots
    fig, ax1 = plt.subplots(figsize=figsize)
    #Set other side y-axis for lookback time scalings
    ax2 = ax1.twinx()

    #Set axis scales based on what data sampling we used 
    if yaxis_type == 'lin' and xaxis_type == 'log':
        CS1 = ax1.contourf(np.log10(sample_x),sample_y,logSNR,logLevels,cmap = colormap)
        ax2.contour(np.log10(sample_x),sample_y,logSNR,print_logLevels,colors = 'k',alpha=1.0)
        ax1.set_xlim(np.log10(xlabel_min),np.log10(xlabel_max))
        ax1.set_ylim(ylabel_min,ylabel_max)
        x_labels = np.logspace(np.log10(xlabel_min),np.log10(xlabel_max),np.log10(xlabel_max)-np.log10(xlabel_min)+1)
        y_labels = np.linspace(ylabel_min,ylabel_max,ylabel_max-ylabel_min+1)
        ax1.set_yticks(y_labels)
        ax1.set_xticks(np.log10(x_labels))
    elif yaxis_type == 'log' and xaxis_type == 'lin':
        CS1 = ax1.contourf(sample_x,np.log10(sample_y),logSNR,logLevels,cmap = colormap)
        ax2.contour(sample_x,np.log10(sample_y),logSNR,print_logLevels,colors = 'k',alpha=1.0)
        ax1.set_xlim(xlabel_min,xlabel_max)
        ax1.set_ylim(np.log10(ylabel_min),np.log10(ylabel_max))
        x_labels = np.linspace(xlabel_min,xlabel_max,xlabel_max-xlabel_min+1)
        y_labels = np.logspace(np.log10(ylabel_min),np.log10(ylabel_max),np.log10(ylabel_max)-np.log10(ylabel_min)+1)
        ax1.set_xticks(x_labels)
        ax1.set_yticks(np.log10(y_labels))
    else:
        CS1 = ax1.contourf(np.log10(sample_x),np.log10(sample_y),logSNR,logLevels,cmap = colormap)
        ax2.contour(np.log10(sample_x),np.log10(sample_y),logSNR,print_logLevels,colors = 'k',alpha=1.0)
        ax1.set_xlim(np.log10(xlabel_min),np.log10(xlabel_max))
        ax1.set_ylim(np.log10(ylabel_min),np.log10(ylabel_max))
        x_labels = np.logspace(np.log10(xlabel_min),np.log10(xlabel_max),np.log10(xlabel_max)-np.log10(xlabel_min)+1)
        y_labels = np.logspace(np.log10(ylabel_min),np.log10(ylabel_max),np.log10(ylabel_max)-np.log10(ylabel_min)+1)
        ax1.set_yticks(np.log10(y_labels))
        ax1.set_xticks(np.log10(x_labels))

    #Set axes labels and whether log or linearly spaced
    if var_x == 'M':
        ax1.set_xlabel(r'$M_{\rm tot}$ $[M_{\odot}]$',fontsize = labelsize)
        ax1.set_xticklabels([r'$10^{%i}$' %x if int(x) > 1 else r'$%i$' %(10**x) for x in np.log10(x_labels)],fontsize = axissize)
    elif var_x == 'q':
        ax1.set_xlabel(r'$q$',fontsize = labelsize)
        ax1.set_xticklabels(x_labels,fontsize = axissize,rotation=45)
    elif var_x == 'z':
        ax1.set_xlabel(r'${\rm Redshift}$',fontsize = labelsize)
        ax1.set_xticklabels([x if int(x) < 1 else int(x) for x in x_labels],fontsize = axissize)
    elif var_x == 'chi1' or var_x == 'chi2':
        x_labels = np.arange(round(xlabel_min*10),round(xlabel_max*10)+1,1)/10
        ax1.set_xticks(x_labels)
        ax1.set_xlabel(r'${\rm Spin}$',fontsize = labelsize)
        ax1.set_xticklabels(x_labels,fontsize = axissize,rotation=45)
        ax1.legend(loc='lower right')
    elif var_x == 'L':
        ax1.axvline(x=np.log10(2.5*u.Gm.to('m')),linestyle='--',color='k',label='Proposed arm length')
        ax1.set_xlabel(r'${\rm Armlength}$ $[m]$',fontsize = labelsize)
        ax1.set_xticklabels([r'$10^{%i}$' %x if int(x) > 1 else r'$%i$' %(10**x) for x in np.log10(x_labels)],fontsize = axissize)
    elif var_x == 'T_obs':
        ax1.set_xlabel(r'${\rm T_{obs}}$ $[yr]$',fontsize = labelsize)
        ax1.set_xticklabels([r'$%i$' %int(x) for x in x_labels/u.yr.to('s')],fontsize = axissize)

    if var_y == 'M':
        ax1.set_ylabel(r'$M_{\rm tot}$ $[M_{\odot}]$',fontsize = labelsize)
        ax1.set_yticklabels([r'$10^{%i}$' %y if int(y) > 1 else r'$%i$' %(10**y) for y in np.log10(y_labels)],fontsize = axissize)
    elif var_y == 'q':
        ax1.set_ylabel(r'$q$',fontsize = labelsize)
        ax1.set_yticklabels(y_labels,fontsize = axissize,rotation=45)
    elif var_y == 'z':
        ax1.set_ylabel(r'${\rm Redshift}$',fontsize = labelsize)
        ax1.set_yticklabels([y if int(y) < 1 else int(y) for y in y_labels],\
            fontsize = axissize)
    elif var_y == 'chi1' or var_y == 'chi2':
        y_labels = np.arange(round(ylabel_min*10),round(ylabel_max*10)+1,1)/10
        ax1.set_yticks(y_labels)
        ax1.set_ylabel(r'${\rm Spin}$',fontsize = labelsize)
        ax1.set_yticklabels(y_labels,fontsize = axissize,rotation=45)
    elif var_y == 'L':
        ax1.axhline(y=np.log10(2.5*u.Gm.to('m')),linestyle='--',color='k',label='Proposed arm length')
        ax1.set_ylabel(r'${\rm Armlength}$ $[m]$',fontsize = labelsize)
        ax1.set_yticklabels([r'$10^{%i}$' %y if int(y) > 1 else r'$%i$' %(10**y) for y in np.log10(y_labels)],fontsize = axissize)
        ax1.legend(loc='lower right')
    elif var_y == 'T_obs':
        ax1.set_ylabel(r'${\rm T_{obs}}$ $[yr]$',fontsize = labelsize)
        ax1.set_yticklabels([r'$%i$' %int(y) for y in y_labels/u.yr.to('s')],fontsize = axissize)

    ax1.yaxis.set_label_coords(-.10,.5)

    #If true, display luminosity distance on right side of plot
    if dl_axis:
        dists_min = cosmo.luminosity_distance(source.var_dict['z']['min']).to('Gpc')
        dists_min = np.ceil(np.log10(dists_min.value))
        dists_max = cosmo.luminosity_distance(source.var_dict['z']['max']).to('Gpc')
        dists_max = np.ceil(np.log10(dists_max.value))
        dists = np.arange(dists_min,dists_max)
        dists = 10**dists*u.Gpc

        distticks = [z_at_value(cosmo.luminosity_distance,dist) for dist in dists]
        #Set other side y-axis for lookback time scalings
        ax2.set_yticks(np.log10(distticks))
        #ax2.set_yticklabels(['%f' %dist for dist in distticks],fontsize = axissize)
        ax2.set_yticklabels([r'$10^{%i}$' %np.log10(dist) if np.abs(int(np.log10(dist))) > 1 else '{:g}'.format(dist) for dist in dists.value],fontsize = axissize)
        ax2.set_ylabel(r'$D_{L}$ [Gpc]',fontsize=labelsize)
        cbar = fig.colorbar(CS1,ax=(ax1,ax2),pad=0.01,ticks=print_logLevels)
    else:
        #Make colorbar
        cbar = fig.colorbar(CS1,ticks=print_logLevels)
        #Remove y-axis labels
        ax2.tick_params(axis='y',right=False,labelright=False)
    cbar.set_label(r'$SNR$',fontsize = labelsize)
    cbar.ax.tick_params(labelsize = axissize)
    cbar.ax.set_yticklabels([r'$10^{%i}$' %x if int(x) > 1 else r'$%i$' %(10**x) for x in print_logLevels])

    if display:
        plt.show()

    #Save Figure to File
    if figloc != None:
        fig.savefig(figloc,bbox_inches='tight')